from io import StringIO
import sys
from unittest import TestCase
import numpy as np
import os.path
from redisai import Client
from ml2rt import load_model
from redis.exceptions import ResponseError


DEBUG = False


class Capturing(list):
    def __enter__(self):
        self._stdout = sys.stdout
        sys.stdout = self._stringio = StringIO()
        return self

    def __exit__(self, *args):
        self.extend(self._stringio.getvalue().splitlines())
        del self._stringio    # free up some memory
        sys.stdout = self._stdout


MODEL_DIR = os.path.dirname(os.path.abspath(__file__)) + '/testdata'
script = r"""
def bar(a, b):
    return a + b
"""

class RedisAITestBase(TestCase):
    def setUp(self):
        super().setUp()
        self.get_client().flushall()

    def get_client(self, debug=DEBUG):
        return Client(debug)


class ClientTestCase(RedisAITestBase):

    def test_set_non_numpy_tensor(self):
        con = self.get_client()
        con.tensorset('x', (2, 3, 4, 5), dtype='float')
        result = con.tensorget('x', as_numpy=False)
        self.assertEqual([2, 3, 4, 5], result['values'])
        self.assertEqual([4], result['shape'])

        con.tensorset('x', (2, 3, 4, 5), dtype='float64')
        result = con.tensorget('x', as_numpy=False)
        self.assertEqual([2, 3, 4, 5], result['values'])
        self.assertEqual([4], result['shape'])
        self.assertEqual('DOUBLE', result['dtype'])

        con.tensorset('x', (2, 3, 4, 5), dtype='int16', shape=(2, 2))
        result = con.tensorget('x', as_numpy=False)
        self.assertEqual([2, 3, 4, 5], result['values'])
        self.assertEqual([2, 2], result['shape'])

        with self.assertRaises(TypeError):
            con.tensorset('x', (2, 3, 4, 5), dtype='wrongtype', shape=(2, 2))
        con.tensorset('x', (2, 3, 4, 5), dtype='int8', shape=(2, 2))
        result = con.tensorget('x', as_numpy=False)
        self.assertEqual('INT8', result['dtype'])
        self.assertEqual([2, 3, 4, 5], result['values'])
        self.assertEqual([2, 2], result['shape'])
        self.assertIn('values', result)

        with self.assertRaises(TypeError):
            con.tensorset('x')
            con.tensorset(1)

    def test_tensorget_meta(self):
        con = self.get_client()
        con.tensorset('x', (2, 3, 4, 5), dtype='float')
        result = con.tensorget('x', meta_only=True)
        self.assertNotIn('values', result)
        self.assertEqual([4], result['shape'])

    def test_numpy_tensor(self):
        con = self.get_client()

        input_array = np.array([2, 3], dtype=np.float32)
        con.tensorset('x', input_array)
        values = con.tensorget('x')
        self.assertEqual(values.dtype, np.float32)

        input_array = np.array([2, 3], dtype=np.float64)
        con.tensorset('x', input_array)
        values = con.tensorget('x')
        self.assertEqual(values.dtype, np.float64)

        input_array = np.array([2, 3])
        con.tensorset('x', input_array)
        values = con.tensorget('x')

        self.assertTrue(np.allclose([2, 3], values))
        self.assertEqual(values.dtype, np.int64)
        self.assertEqual(values.shape, (2,))
        self.assertTrue((np.allclose(values, input_array)))
        ret = con.tensorset('x', values)
        self.assertEqual(ret, 'OK')

        stringarr = np.array('dummy')
        with self.assertRaises(TypeError):
            con.tensorset('trying', stringarr)

    def test_modelget_meta(self):
        model_path = os.path.join(MODEL_DIR, 'graph.pb')
        model_pb = load_model(model_path)
        con = self.get_client()
        con.modelset('m', 'tf', 'cpu', model_pb,
                     inputs=['a', 'b'], outputs=['mul'], tag='v1.0')
        model = con.modelget('m', meta_only=True)
        self.assertEqual(model, {'backend': 'TF', 'device': 'cpu', 'tag': 'v1.0'})

    def test_modelrun_non_list_input_output(self):
        model_path = os.path.join(MODEL_DIR, 'graph.pb')
        model_pb = load_model(model_path)
        con = self.get_client()
        con.modelset('m', 'tf', 'cpu', model_pb,
                     inputs=['a', 'b'], outputs=['mul'], tag='v1.0')
        con.tensorset('a', (2, 3), dtype='float')
        con.tensorset('b', (2, 3), dtype='float')
        ret = con.modelrun('m', ['a', 'b'], 'out')
        self.assertEqual(ret, 'OK')

    def test_nonasciichar(self):
        nonascii = 'ĉ'
        model_path = os.path.join(MODEL_DIR, 'graph.pb')
        model_pb = load_model(model_path)
        con = self.get_client()
        con.modelset('m' + nonascii, 'tf', 'cpu', model_pb,
                     inputs=['a', 'b'], outputs=['mul'], tag='v1.0')
        con.tensorset('a' + nonascii, (2, 3), dtype='float')
        con.tensorset('b', (2, 3), dtype='float')
        con.modelrun('m' + nonascii, ['a' + nonascii, 'b'], ['c' + nonascii])
        tensor = con.tensorget('c' + nonascii)
        self.assertTrue((np.allclose(tensor, [4., 9.])))

    def test_run_tf_model(self):
        model_path = os.path.join(MODEL_DIR, 'graph.pb')
        bad_model_path = os.path.join(MODEL_DIR, 'pt-minimal.pt')

        model_pb = load_model(model_path)
        wrong_model_pb = load_model(bad_model_path)

        con = self.get_client()
        con.modelset('m', 'tf', 'cpu', model_pb,
                     inputs=['a', 'b'], outputs=['mul'], tag='v1.0')
        con.modeldel('m')
        self.assertRaises(ResponseError, con.modelget, 'm')
        con.modelset('m', 'tf', 'cpu', model_pb,
                     inputs=['a', 'b'], outputs='mul', tag='v1.0')

        # wrong model
        self.assertRaises(ResponseError,
                          con.modelset, 'm', 'tf', 'cpu',
                          wrong_model_pb,
                          inputs=['a', 'b'], outputs=['mul'])
        # missing inputs/outputs
        self.assertRaises(ValueError,
                          con.modelset, 'm', 'tf', 'cpu',
                          wrong_model_pb)

        # wrong backend
        self.assertRaises(ResponseError,
                          con.modelset, 'm', 'torch', 'cpu',
                          model_pb,
                          inputs=['a', 'b'], outputs=['mul'])

        con.tensorset('a', (2, 3), dtype='float')
        con.tensorset('b', (2, 3), dtype='float')
        con.modelrun('m', ['a', 'b'], ['c'])
        tensor = con.tensorget('c')
        self.assertTrue(np.allclose([4, 9], tensor))
        model_det = con.modelget('m')
        self.assertTrue(model_det['backend'] == 'TF')
        self.assertTrue(model_det['device'] == 'cpu')  # TODO; RedisAI returns small letter
        self.assertTrue(model_det['tag'] == 'v1.0')
        con.modeldel('m')
        self.assertRaises(ResponseError, con.modelget, 'm')

    def test_scripts(self):
        con = self.get_client()
        self.assertRaises(ResponseError, con.scriptset,
                          'ket', 'cpu', 'return 1')
        con.scriptset('ket', 'cpu', script)
        con.tensorset('a', (2, 3), dtype='float')
        con.tensorset('b', (2, 3), dtype='float')
        # try with bad arguments:
        self.assertRaises(ResponseError,
                          con.scriptrun, 'ket', 'bar', inputs=['a'], outputs=['c'])
        con.scriptrun('ket', 'bar', inputs=['a', 'b'], outputs=['c'])
        tensor = con.tensorget('c', as_numpy=False)
        self.assertEqual([4, 6], tensor['values'])
        script_det = con.scriptget('ket')
        self.assertTrue(script_det['device'] == 'cpu')
        self.assertTrue(script_det['source'] == script)
        script_det = con.scriptget('ket', meta_only=True)
        self.assertTrue(script_det['device'] == 'cpu')
        self.assertNotIn('source', script_det)
        con.scriptdel('ket')
        self.assertRaises(ResponseError, con.scriptget, 'ket')

    def test_run_onnxml_model(self):
        mlmodel_path = os.path.join(MODEL_DIR, 'boston.onnx')
        onnxml_model = load_model(mlmodel_path)
        con = self.get_client()
        con.modelset("onnx_model", 'onnx', 'cpu', onnxml_model)
        tensor = np.ones((1, 13)).astype(np.float32)
        con.tensorset("input", tensor)
        con.modelrun("onnx_model", ["input"], ["output"])
        # tests `convert_to_num`
        outtensor = con.tensorget("output", as_numpy=False)
        self.assertEqual(int(float(outtensor['values'][0])), 24)

    def test_run_onnxdl_model(self):
        # A PyTorch model that finds the square
        dlmodel_path = os.path.join(MODEL_DIR, 'findsquare.onnx')
        onnxdl_model = load_model(dlmodel_path)
        con = self.get_client()
        con.modelset("onnx_model", 'onnx', 'cpu', onnxdl_model)
        tensor = np.array((2,)).astype(np.float32)
        con.tensorset("input", tensor)
        con.modelrun("onnx_model", ["input"], ["output"])
        outtensor = con.tensorget("output")
        self.assertTrue(np.allclose(outtensor, [4.0]))

    def test_run_pytorch_model(self):
        model_path = os.path.join(MODEL_DIR, 'pt-minimal.pt')
        ptmodel = load_model(model_path)
        con = self.get_client()
        con.modelset("pt_model", 'torch', 'cpu', ptmodel, tag='v1.0')
        con.tensorset('a', [2, 3, 2, 3], shape=(2, 2), dtype='float')
        con.tensorset('b', [2, 3, 2, 3], shape=(2, 2), dtype='float')
        con.modelrun("pt_model", ["a", "b"], ["output"])
        output = con.tensorget('output', as_numpy=False)
        self.assertTrue(np.allclose(output['values'], [4, 6, 4, 6]))

    def test_run_tflite_model(self):
        model_path = os.path.join(MODEL_DIR, 'mnist_model_quant.tflite')
        tflmodel = load_model(model_path)
        con = self.get_client()
        con.modelset("tfl_model", 'tflite', 'cpu', tflmodel)
        img = np.random.random((1, 1, 28, 28)).astype(np.float)
        con.tensorset('img', img)
        con.modelrun("tfl_model", ["img"], ["output1", "output2"])
        output = con.tensorget('output1')
        self.assertTrue(np.allclose(output, [8]))

    def test_info(self):
        model_path = os.path.join(MODEL_DIR, 'graph.pb')
        model_pb = load_model(model_path)
        con = self.get_client()
        con.modelset('m', 'tf', 'cpu', model_pb,
                     inputs=['a', 'b'], outputs=['mul'])
        first_info = con.infoget('m')
        expected = {'key': 'm', 'type': 'MODEL', 'backend': 'TF', 'device': 'cpu',
                    'tag': '', 'duration': 0, 'samples': 0, 'calls': 0, 'errors': 0}
        self.assertEqual(first_info, expected)
        con.tensorset('a', (2, 3), dtype='float')
        con.tensorset('b', (2, 3), dtype='float')
        con.modelrun('m', ['a', 'b'], ['c'])
        con.modelrun('m', ['a', 'b'], ['c'])
        second_info = con.infoget('m')
        self.assertEqual(second_info['calls'], 2)  # 2 model runs
        con.inforeset('m')
        third_info = con.infoget('m')
        self.assertEqual(first_info, third_info)  # before modelrun and after reset

    def test_model_scan(self):
        model_path = os.path.join(MODEL_DIR, 'graph.pb')
        model_pb = load_model(model_path)
        con = self.get_client()
        con.modelset('m', 'tf', 'cpu', model_pb,
                     inputs=['a', 'b'], outputs=['mul'], tag='v1.2')
        model_path = os.path.join(MODEL_DIR, 'pt-minimal.pt')
        ptmodel = load_model(model_path)
        con = self.get_client()
        con.modelset("pt_model", 'torch', 'cpu', ptmodel)
        mlist = con.modelscan() # TODO: modelscan issues in RedisAI
        self.assertEqual(mlist, [['pt_model', ''], ['m', 'v1.2']])

    def test_script_scan(self):
        con = self.get_client()
        con.scriptset('ket1', 'cpu', script, tag='v1.0')
        con.scriptset('ket2', 'cpu', script)
        slist = con.scriptscan()
        self.assertEqual(slist, [['ket1', 'v1.0'], ['ket2', '']])

    def test_debug(self):
        con = self.get_client(debug=True)
        with Capturing() as output:
            con.tensorset('x', (2, 3, 4, 5), dtype='float')
        self.assertEqual(['AI.TENSORSET x FLOAT 4 VALUES 2 3 4 5'], output)


class DagTestCase(RedisAITestBase):
    def setUp(self):
        super().setUp()
        con = self.get_client()
        model_path = os.path.join(MODEL_DIR, 'pt-minimal.pt')
        ptmodel = load_model(model_path)
        con.modelset("pt_model", 'torch', 'cpu', ptmodel, tag='v1.0')

    def test_dagrun_with_load(self):
        con = self.get_client()
        con.tensorset('a', [2, 3, 2, 3], shape=(2, 2), dtype='float')

        dag = con.dag(load='a')
        dag.tensorset('b', [2, 3, 2, 3], shape=(2, 2), dtype='float')
        dag.modelrun("pt_model", ["a", "b"], ["output"])
        dag.tensorget('output')
        result = dag.run()
        expected = ['OK', 'OK', np.array([[4., 6.], [4., 6.]], dtype=np.float32)]
        self.assertTrue(np.allclose(expected.pop(), result.pop()))
        self.assertEqual(expected, result)
        self.assertRaises(ResponseError, con.tensorget, 'b')

    def test_dagrun_with_persist(self):
        con = self.get_client()

        dag = con.dag(persist='wrongkey')  # this won't raise Error
        dag.tensorset('a', [2, 3, 2, 3], shape=(2, 2), dtype='float').run()

        dag = con.dag(persist=['b'])
        dag.tensorset('a', [2, 3, 2, 3], shape=(2, 2), dtype='float')
        dag.tensorset('b', [2, 3, 2, 3], shape=(2, 2), dtype='float')
        dag.tensorget('b')
        result = dag.run()
        b = con.tensorget('b')
        self.assertTrue(np.allclose(b, result[-1]))
        self.assertEqual(b.dtype, np.float32)
        self.assertEqual(len(result), 3)

    def test_dagrun_calling_on_return(self):
        con = self.get_client()
        con.tensorset('a', [2, 3, 2, 3], shape=(2, 2), dtype='float')
        result = con.\
            dag(load='a').\
            tensorset('b', [2, 3, 2, 3], shape=(2, 2), dtype='float').\
            modelrun("pt_model", ["a", "b"], ["output"]).\
            tensorget('output').\
            run()
        expected = ['OK', 'OK', np.array([[4., 6.], [4., 6.]], dtype=np.float32)]
        self.assertTrue(np.allclose(expected.pop(), result.pop()))
        self.assertEqual(expected, result)

    def test_dagrun_without_load_and_persist(self):
        con = self.get_client()

        dag = con.dag(load='wrongkey')
        with self.assertRaises(ResponseError):
            dag.tensorget('wrongkey').run()

        dag = con.dag()
        dag.tensorset('a', [2, 3, 2, 3], shape=(2, 2), dtype='float')
        dag.tensorset('b', [2, 3, 2, 3], shape=(2, 2), dtype='float')
        dag.modelrun("pt_model", ["a", "b"], ["output"])
        dag.tensorget('output')
        result = dag.run()
        expected = ['OK', 'OK', 'OK', np.array([[4., 6.], [4., 6.]], dtype=np.float32)]
        self.assertTrue(np.allclose(expected.pop(), result.pop()))
        self.assertEqual(expected, result)

    def test_dagrun_with_load_and_persist(self):
        con = self.get_client()
        con.tensorset('a', [2, 3, 2, 3], shape=(2, 2), dtype='float')
        con.tensorset('b', [2, 3, 2, 3], shape=(2, 2), dtype='float')
        dag = con.dag(load=['a', 'b'], persist='output')
        dag.modelrun("pt_model", ["a", "b"], ["output"])
        dag.tensorget('output')
        result = dag.run()
        expected = ['OK', np.array([[4., 6.], [4., 6.]], dtype=np.float32)]
        result_outside_dag = con.tensorget('output')
        self.assertTrue(np.allclose(expected.pop(), result.pop()))
        result = dag.run()
        self.assertTrue(np.allclose(result_outside_dag, result.pop()))
        self.assertEqual(expected, result)

    def test_dagrunRO(self):
        con = self.get_client()
        con.tensorset('a', [2, 3, 2, 3], shape=(2, 2), dtype='float')
        con.tensorset('b', [2, 3, 2, 3], shape=(2, 2), dtype='float')
        with self.assertRaises(RuntimeError):
            con.dag(load=['a', 'b'], persist='output', readonly=True)
        dag = con.dag(load=['a', 'b'], readonly=True)
        dag.modelrun("pt_model", ["a", "b"], ["output"])
        dag.tensorget('output')
        result = dag.run()
        expected = ['OK', np.array([[4., 6.], [4., 6.]], dtype=np.float32)]
        self.assertTrue(np.allclose(expected.pop(), result.pop()))





from functools import wraps, partial
from typing import Union, AnyStr, ByteString, List, Sequence, Any
import warnings

from redis import StrictRedis
import numpy as np

from . import utils
from .command_builder import Builder


builder = Builder()


def enable_debug(f):
    @wraps(f)
    def wrapper(*args):
        print(*args)
        return f(*args)
    return wrapper


class Dag:
    def __init__(self, load, persist, executor, readonly=False):
        self.result_processors = []
        if readonly:
            if persist:
                raise RuntimeError("READONLY requests cannot write (duh!) and should not "
                                   "have PERSISTing values")
            self.commands = ['AI.DAGRUN_RO']
        else:
            self.commands = ['AI.DAGRUN']
        if load:
            if not isinstance(load, (list, tuple)):
                self.commands += ["LOAD", 1, load]
            else:
                self.commands += ["LOAD", len(load), *load]
        if persist:
            if not isinstance(persist, (list, tuple)):
                self.commands += ["PERSIST", 1, persist, '|>']
            else:
                self.commands += ["PERSIST", len(persist), *persist, '|>']
        elif load:
            self.commands.append('|>')
        self.executor = executor

    def tensorset(self,
                  key: AnyStr,
                  tensor: Union[np.ndarray, list, tuple],
                  shape: Sequence[int] = None,
                  dtype: str = None) -> Any:
        args = builder.tensorset(key, tensor, shape, dtype)
        self.commands.extend(args)
        self.commands.append("|>")
        self.result_processors.append(bytes.decode)
        return self

    def tensorget(self,
                  key: AnyStr, as_numpy: bool = True,
                  meta_only: bool = False) -> Any:
        args = builder.tensorget(key, as_numpy, meta_only)
        self.commands.extend(args)
        self.commands.append("|>")
        self.result_processors.append(partial(utils.tensorget_postprocessor,
                                              as_numpy,
                                              meta_only))
        return self

    def modelrun(self,
                 key: AnyStr,
                 inputs: Union[AnyStr, List[AnyStr]],
                 outputs: Union[AnyStr, List[AnyStr]]) -> Any:
        args = builder.modelrun(key, inputs, outputs)
        self.commands.extend(args)
        self.commands.append("|>")
        self.result_processors.append(bytes.decode)
        return self

    def run(self):
        results = self.executor(*self.commands)
        out = []
        for res, fn in zip(results, self.result_processors):
            out.append(fn(res))
        return out


class Client(StrictRedis):
    """
    Redis client build specifically for the RedisAI module. It takes all the necessary
    parameters to establish the connection and an optional ``debug`` parameter on
    initialization

    Parameters
    ----------

    debug : bool
        If debug mode is ON, then each command that is sent to the server is
        printed to the terminal

    Example
    -------
    >>> from redisai import Client
    >>> con = Client(host='localhost', port=6379)
    """
    def __init__(self, debug=False, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if debug:
            self.execute_command = enable_debug(super().execute_command)

    def dag(self, load: Sequence = None, persist: Sequence = None,
            readonly: bool = False) -> Dag:
        """ It returns a DAG object on which other DAG-allowed operations can be called. For
        more details about DAG in RedisAI, refer to the RedisAI documentation.

        Parameters
        ----------
        load : Union[AnyStr, List[AnyStr]]
            Load the list of given values from the keyspace to DAG scope
        persist : Union[AnyStr, List[AnyStr]]
            Write the list of given key, values to the keyspace from DAG scope
        readonly : bool
            If True, it triggers AI.DAGRUN_RO, the read only DAG which cannot write (PERSIST) to
            the keyspace. But since it can't write, it can execute on replicas


        Returns
        -------
        Any
            Dag object which holds other operations permitted inside DAG as attributes

        Example
        -------
        >>> con.tensorset('tensor', ...)
        'OK'
        >>> con.modelset('model', ...)
        'OK'
        >>> dag = con.dag(load=['tensor'], persist=['output'])
        >>> dag.tensorset('another', ...)
        >>> dag.modelrun('model', inputs=['tensor', 'another'], outputs=['output'])
        >>> output = dag.tensorget('output').run()
        >>> # You can even chain the operations
        >>> result = dag.tensorset(**akwargs).modelrun(**bkwargs).tensorget(**ckwargs).run()
        """
        return Dag(load, persist, self.execute_command, readonly)

    def loadbackend(self, identifier: AnyStr, path: AnyStr) -> str:
        """
        RedisAI by default won't load any backends. User can either explicitly
        load the backend by using this function or let RedisAI load the required
        backend from the default path on-demand.

        Parameters
        ----------
        identifier : str
            Representing which backend. Allowed values - TF, TFLITE, TORCH & ONNX
        path: str
            Path to the shared object of the backend

        Returns
        -------
        str
            'OK' if success, raise an exception otherwise

        Example
        -------
        >>> con.loadbackend('TORCH', '/path/to/the/backend/redisai_torch.so')
        'OK'
        """
        args = builder.loadbackend(identifier, path)
        return self.execute_command(*args).decode()

    def modelset(self,
                 key: AnyStr,
                 backend: str,
                 device: str,
                 data: ByteString,
                 batch: int = None,
                 minbatch: int = None,
                 tag: AnyStr = None,
                 inputs: Union[AnyStr, List[AnyStr]] = None,
                 outputs: Union[AnyStr, List[AnyStr]] = None) -> str:
        """
        Set the model on provided key.

        Parameters
        ----------
        key : AnyStr
            Key name
        backend : str
            Backend name. Allowed backends are TF, TORCH, TFLITE, ONNX
        device : str
            Device name. Allowed devices are CPU and GPU. If multiple GPUs are available,
            it can be specified using the format GPU:<gpu number>. For example: GPU:0
        data : bytes
            Model graph read as bytes string
        batch : int
            Number of batches for doing auto-batching
        minbatch : int
            Minimum number of samples required in a batch for model execution
        tag : AnyStr
            Any string that will be saved in RedisAI as tag for the model
        inputs : Union[AnyStr, List[AnyStr]]
            Input node(s) in the graph. Required only Tensorflow graphs
        outputs : Union[AnyStr, List[AnyStr]]
            Output node(s) in the graph Required only for Tensorflow graphs

        Returns
        -------
        str
            'OK' if success, raise an exception otherwise

        Example
        -------
        >>> # Torch model
        >>> model_path = os.path.join('path/to/TorchScriptModel.pt')
        >>> model = open(model_path, 'rb').read()
        >>> con.modelset("model", 'torch', 'cpu', model, tag='v1.0')
        'OK'
        >>> # Tensorflow model
        >>> model_path = os.path.join('/path/to/tf_frozen_graph.pb')
        >>> model = open(model_path, 'rb').read()
        >>> con.modelset('m', 'tf', 'cpu', model,
        ...              inputs=['a', 'b'], outputs=['mul'], tag='v1.0')
        'OK'
        """
        args = builder.modelset(key, backend, device, data,
                                batch, minbatch, tag, inputs, outputs)
        return self.execute_command(*args).decode()

    def modelget(self, key: AnyStr, meta_only=False) -> dict:
        """
        Fetch the model details and the model blob back from RedisAI

        Parameters
        ----------
        key : AnyStr
            Model key in RedisAI
        meta_only : bool
            If True, only the meta data will be fetched, not the model blob

        Returns
        -------
        dict
            A dictionary of model details such as device, backend etc. The model
            blob will be available at the key 'blob'

        Example
        -------
        >>> con.modelget('model', meta_only=True)
        {'backend': 'TF', 'device': 'cpu', 'tag': 'v1.0'}
        """
        args = builder.modelget(key, meta_only)
        rv = self.execute_command(*args)
        return utils.list2dict(rv)

    def modeldel(self, key: AnyStr) -> str:
        """
        Delete the model from the RedisAI server

        Parameters
        ----------
        key : AnyStr
            Key of the model to be deleted

        Returns
        -------
        str
            'OK' if success, raise an exception otherwise

        Example
        -------
        >>> con.modeldel('model')
        'OK'
        """
        args = builder.modeldel(key)
        return self.execute_command(*args).decode()

    def modelrun(self,
                 key: AnyStr,
                 inputs: Union[AnyStr, List[AnyStr]],
                 outputs: Union[AnyStr, List[AnyStr]]) -> str:
        """
        Run the model using input(s) which are already in the scope and are associated
        to some keys. Modelrun also needs the output key name(s) to store the output
        from the model. The number of outputs from the model and the number of keys
        provided here must be same. Otherwise, RedisAI throws an error

        Parameters
        ----------
        key : str
            Model key to run
        inputs : Union[AnyStr, List[AnyStr]]
            Tensor(s) which is already saved in the RedisAI using a tensorset call. These
            tensors will be used as the input for the modelrun
        outputs : Union[AnyStr, List[AnyStr]]
            keys on which the outputs to be saved. If those keys exist already, modelrun
            will overwrite them with new values

        Returns
        -------
        str
            'OK' if success, raise an exception otherwise

        Example
        -------
        >>> con.modelset('m', 'tf', 'cpu', model_pb,
        ...              inputs=['a', 'b'], outputs=['mul'], tag='v1.0')
        'OK'
        >>> con.tensorset('a', (2, 3), dtype='float')
        'OK'
        >>> con.tensorset('b', (2, 3), dtype='float')
        'OK'
        >>> con.modelrun('m', ['a', 'b'], ['c'])
        'OK'
        """
        args = builder.modelrun(key, inputs, outputs)
        return self.execute_command(*args).decode()

    def modelscan(self) -> List[List[AnyStr]]:
        """
        Returns the list of all the models in the RedisAI server. Modelscan API is
        currently experimental and might be removed or changed in the future without
        warning

        Returns
        -------
        List[List[AnyStr]]
            List of list of models and tags for each model if they existed

        Example
        -------
        >>> con.modelscan()
        [['pt_model', ''], ['m', 'v1.2']]
        """
        warnings.warn("Experimental: Model List API is experimental and might change "
                      "in the future without any notice", UserWarning)
        args = builder.modelscan()
        result = self.execute_command(*args)
        return utils.recursive_bytetransform(result, lambda x: x.decode())

    def tensorset(self,
                  key: AnyStr,
                  tensor: Union[np.ndarray, list, tuple],
                  shape: Sequence[int] = None,
                  dtype: str = None) -> str:
        """
        Set the tensor to a key in RedisAI

        Parameters
        ----------
        key : AnyStr
            The name of the tensor
        tensor : Union[np.ndarray, list, tuple]
            A `np.ndarray` object or Python list or tuple
        shape : Sequence[int]
            Shape of the tensor. Required if `tensor` is list or tuple
        dtype : str
            Data type of the tensor. Required if `tensor` is list or tuple

        Returns
        -------
        str
            'OK' if success, raise an exception otherwise

        Example
        -------
        >>> con.tensorset('a', (2, 3), dtype='float')
        'OK'
        >>> input_array = np.array([2, 3], dtype=np.float32)
        >>> con.tensorset('x', input_array)
        'OK'
        """
        args = builder.tensorset(key, tensor, shape, dtype)
        return self.execute_command(*args).decode()

    def tensorget(self,
                  key: AnyStr, as_numpy: bool = True,
                  meta_only: bool = False) -> Union[dict, np.ndarray]:
        """
        Retrieve the value of a tensor from the server. By default it returns the numpy
        array but it can be controlled using the `as_type` and `meta_only` argument.

        Parameters
        ----------
        key : AnyStr
            The name of the tensor
        as_numpy : bool
            If True, returns a numpy.ndarray. Returns the value as a list and the
            metadata in a dictionary if False. This flag also decides how to fetch
            the value from the RedisAI server, which also has performance implications
        meta_only : bool
            If True, the value is not retrieved, only the shape and the type

        Returns
        -------
        Union[dict, np.ndarray]
            Returns a dictionary of data or a numpy array. Default is numpy array

        Example
        -------
        >>> con.tensorget('x')
        array([2, 3, 4])
        >>> con.tensorget('x' as_numpy=False)
        {'values': [2, 3, 4], 'dtype': 'INT64', 'shape': [3]}
        >>> con.tensorget('x', meta_only=True)
        {'dtype': 'INT64', 'shape': [3]}
        """
        args = builder.tensorget(key, as_numpy, meta_only)
        res = self.execute_command(*args)
        return utils.tensorget_postprocessor(as_numpy, meta_only, res)

    def scriptset(self, key: AnyStr, device: str, script: str, tag: AnyStr = None) -> str:
        """
        Set the script to RedisAI. Action similar to Modelset. RedisAI uses the TorchScript
        engine to execute the script. So the script should have only TorchScript supported
        constructs. That being said, it's important to mention that using redisai script
        to do post processing or pre processing for a Tensorflow (or any other backend)
        is completely valid. For more details about TorchScript and supported ops,
        checkout TorchScript documentation.

        Parameters
        ----------
        key : AnyStr
            Script key at the server
        device : str
            Device name. Allowed devices are CPU and GPU. If multiple GPUs are available.
            it can be specified using the format GPU:<gpu number>. For example: GPU:0
        script : str
            Script itself, as a Python string
        tag : AnyStr
            Any string that will be saved in RedisAI as tag for the model

        Returns
        -------
        str
            'OK' if success, raise an exception otherwise

        Note
        ----
        Even though ``script`` is pure Python code, it's a subset of Python language and not
        all the Python operations are supported. For more details, checkout TorchScript
        documentation. It's also important to note that that the script is executed on a high
        performance C++ runtime instead of the Python interpreter. And hence ``script`` should
        not have any import statements (A common mistake people make all the time)

        Example
        -------
        >>> script = open(scriptpath).read()
        >>> con.scriptset('ket', 'cpu', script)
        'OK'
        """
        args = builder.scriptset(key, device, script, tag)
        return self.execute_command(*args).decode()

    def scriptget(self, key: AnyStr, meta_only=False) -> dict:
        """
        Get the saved script from RedisAI. Operation similar to model get

        Parameters
        ----------
        key : AnyStr
            Key of the script
        meta_only : bool
            If True, only the meta data will be fetched, not the script itself

        Returns
        -------
        dict
            Dictionary of script details which includes the script at the key ``source``

        Example
        -------
        >>> con.scriptget('ket', meta_only=True)
        {'device': 'cpu'}
        """
        args = builder.scriptget(key, meta_only)
        ret = self.execute_command(*args)
        return utils.list2dict(ret)

    def scriptdel(self, key: AnyStr) -> str:
        """
        Delete the script from the RedisAI server

        Parameters
        ----------
        key : AnyStr
            Script key to be deleted

        Returns
        -------
        str
            'OK' if success, raise an exception otherwise

        Example
        -------
        >>> con.scriptdel('ket')
        'OK'
        """
        args = builder.scriptdel(key)
        return self.execute_command(*args).decode()

    def scriptrun(self,
                  key: AnyStr,
                  function: AnyStr,
                  inputs: Union[AnyStr, Sequence[AnyStr]],
                  outputs: Union[AnyStr, Sequence[AnyStr]]
                  ) -> str:
        """
        Run an already set script. Similar to modelrun

        Parameters
        ----------
        key : AnyStr
            Script key
        function : AnyStr
            Name of the function in the ``script``
        inputs : Union[AnyStr, List[AnyStr]]
            Tensor(s) which is already saved in the RedisAI using a tensorset call. These
            tensors will be used as the input for the modelrun
        outputs : Union[AnyStr, List[AnyStr]]
            keys on which the outputs to be saved. If those keys exist already, modelrun
            will overwrite them with new values

        Returns
        -------
        str
            'OK' if success, raise an exception otherwise

        Example
        -------
        >>> con.scriptrun('ket', 'bar', inputs=['a', 'b'], outputs=['c'])
        'OK'
        """
        args = builder.scriptrun(key, function, inputs, outputs)
        out = self.execute_command(*args)
        return out.decode()

    def scriptscan(self) -> List[List[AnyStr]]:
        """
        Returns the list of all the script in the RedisAI server. Scriptscan API is
        currently experimental and might remove or change in the future without warning

        Returns
        -------
        List[List[AnyStr]]
            List of list of scripts and tags for each script if they existed

        Example
        -------
        >>> con.scriptscan()
        [['ket1', 'v1.0'], ['ket2', '']]
        """
        warnings.warn("Experimental: Script List API is experimental and might change "
                      "in the future without any notice", UserWarning)
        args = builder.scriptscan()
        return utils.recursive_bytetransform(self.execute_command(*args), lambda x: x.decode())

    def infoget(self, key: AnyStr) -> dict:
        """
        Get information such as
        - How long since the model has been running
        - How many samples have been processed
        - How many calls handled
        - How many errors raised
        - etc.

        Parameters
        ----------
        key : AnyStr
            Model key

        Returns
        -------
        dict
            Dictionary of model run details

        Example
        -------
        >>> con.infoget('m')
        {'key': 'm', 'type': 'MODEL', 'backend': 'TF', 'device': 'cpu', 'tag': '',
        'duration': 0, 'samples': 0, 'calls': 0, 'errors': 0}
        """
        args = builder.infoget(key)
        ret = self.execute_command(*args)
        return utils.list2dict(ret)

    def inforeset(self, key: AnyStr) -> str:
        """
        Reset the run information about the model

        Parameters
        ----------
        key : AnyStr
            Model key

        Returns
        -------
        str
            'OK' if success, raise an exception otherwise

        Example
        -------
        >>> con.inforeset('m')
        'OK'
        """
        args = builder.inforeset(key)
        return self.execute_command(*args).decode()



project = 'redisai-py'
copyright = '2020, RedisLabs'
author = 'RedisLabs'
release = '1.0.0'
extensions = ['sphinx.ext.autodoc',
              'sphinx.ext.autosummary',
              'sphinx.ext.extlinks',
              'sphinx.ext.napoleon',
              'sphinx.ext.todo',
              'sphinx.ext.intersphinx',
              "sphinx_rtd_theme"]
templates_path = ['_templates']
source_suffix = '.rst'
master_doc = 'index'
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']
html_theme = 'sphinx_rtd_theme'
html_use_smartypants = True
html_last_updated_fmt = '%b %d, %Y'
html_split_index = False
html_sidebars = {
   '**': ['searchbox.html', 'globaltoc.html', 'sourcelink.html'],
}
html_short_title = '%s-%s' % (project, release)

napoleon_use_ivar = True
napoleon_use_rtype = True
napoleon_use_param = True
napoleon_include_init_with_doc = True

add_module_names = False
doctest_test_doctest_blocks = None
autoclass_content = 'class'



from typing import Union, AnyStr, ByteString, List, Sequence
import numpy as np
from . import utils

# TODO: mypy check


class Builder:
    """ Command Builder class """
    @staticmethod
    def loadbackend(identifier: AnyStr, path: AnyStr) -> Sequence:
        return 'AI.CONFIG LOADBACKEND', identifier, path

    def modelset(self, name: AnyStr, backend: str, device: str, data: ByteString,
                 batch: int, minbatch: int, tag: AnyStr,
                 inputs: Union[AnyStr, List[AnyStr]],
                 outputs: Union[AnyStr, List[AnyStr]]) -> Sequence:
        args = ['AI.MODELSET', name, backend, device]

        if batch is not None:
            args += ['BATCHSIZE', batch]
        if minbatch is not None:
            args += ['MINBATCHSIZE', minbatch]
        if tag is not None:
            args += ['TAG', tag]

        if backend.upper() == 'TF':
            if not(all((inputs, outputs))):
                raise ValueError(
                    'Require keyword arguments input and output for TF models')
            args += ['INPUTS', *utils.listify(inputs)]
            args += ['OUTPUTS', *utils.listify(outputs)]
        chunk_size = 500 * 1024 * 1024
        data_chunks = [data[i:i + chunk_size] for i in range(0, len(data), chunk_size)]
        # TODO: need a test case for this
        args += ['BLOB', *data_chunks]
        return args

    def modelget(self, name: AnyStr, meta_only=False) -> Sequence:
        args = ['AI.MODELGET', name, 'META']
        if not meta_only:
            args.append('BLOB')
        return args

    def modeldel(self, name: AnyStr) -> Sequence:
        return 'AI.MODELDEL', name

    def modelrun(self, name: AnyStr, inputs: List[AnyStr], outputs: List[AnyStr]) -> Sequence:
        args = ('AI.MODELRUN', name, 'INPUTS', *utils.listify(inputs), 'OUTPUTS',
                *utils.listify(outputs))
        return args

    def modelscan(self) -> Sequence:
        return ("AI._MODELSCAN",)

    def tensorset(self,
                  key: AnyStr,
                  tensor: Union[np.ndarray, list, tuple],
                  shape: Sequence[int] = None,
                  dtype: str = None) -> Sequence:
        if np and isinstance(tensor, np.ndarray):
            dtype, shape, blob = utils.numpy2blob(tensor)
            args = ['AI.TENSORSET', key, dtype, *shape, 'BLOB', blob]
        elif isinstance(tensor, (list, tuple)):
            try:
                dtype = utils.dtype_dict[dtype.lower()]
            except KeyError:
                raise TypeError(f'``{dtype}`` is not supported by RedisAI. Currently '
                                f'supported types are {list(utils.dtype_dict.keys())}')
            if shape is None:
                shape = (len(tensor),)
            args = ['AI.TENSORSET', key, dtype, *shape, 'VALUES', *tensor]
        else:
            raise TypeError(f"``tensor`` argument must be a numpy array or a list or a "
                            f"tuple, but got {type(tensor)}")
        return args

    def tensorget(self,
                  key: AnyStr, as_numpy: bool = True,
                  meta_only: bool = False) -> Sequence:
        args = ['AI.TENSORGET', key, 'META']
        if not meta_only:
            if as_numpy is True:
                args.append('BLOB')
            else:
                args.append('VALUES')
        return args

    def scriptset(self, name: AnyStr, device: str, script: str, tag: AnyStr = None) -> Sequence:
        args = ['AI.SCRIPTSET', name, device]
        if tag:
            args += ['TAG', tag]
        args.append("SOURCE")
        args.append(script)
        return args

    def scriptget(self, name: AnyStr, meta_only=False) -> Sequence:
        args = ['AI.SCRIPTGET', name, 'META']
        if not meta_only:
            args.append('SOURCE')
        return args

    def scriptdel(self, name: AnyStr) -> Sequence:
        return 'AI.SCRIPTDEL', name

    def scriptrun(self,
                  name: AnyStr,
                  function: AnyStr,
                  inputs: Union[AnyStr, Sequence[AnyStr]],
                  outputs: Union[AnyStr, Sequence[AnyStr]]
                  ) -> Sequence:
        args = ('AI.SCRIPTRUN', name, function, 'INPUTS', *utils.listify(inputs), 'OUTPUTS',
                *utils.listify(outputs))
        return args

    def scriptscan(self) -> Sequence:
        return ("AI._SCRIPTSCAN",)

    def infoget(self, key: AnyStr) -> Sequence:
        return 'AI.INFO', key

    def inforeset(self, key: AnyStr) -> Sequence:
        return 'AI.INFO', key, 'RESETSTAT'



#!/usr/bin/env python
from setuptools import setup, find_packages


with open('README.rst') as f:
    long_description = f.read()

setup(
    name='redisai',
    version='1.0.0',
    description='RedisAI Python Client',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='http://github.com/RedisAI/redisai-py',
    author='RedisLabs',
    author_email='oss@redislabs.com',
    packages=find_packages(),
    install_requires=['redis', 'hiredis', 'numpy'],
    python_requires='>=3.2',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Topic :: Database',
        'Topic :: Software Development :: Testing'
    ]
)



from typing import Union, ByteString, Sequence, List, AnyStr, Callable
import numpy as np


dtype_dict = {
    'float': 'FLOAT',
    'double': 'DOUBLE',
    'float32': 'FLOAT',
    'float64': 'DOUBLE',
    'int8': 'INT8',
    'int16': 'INT16',
    'int32': 'INT32',
    'int64': 'INT64',
    'uint8': 'UINT8',
    'uint16': 'UINT16',
    'uint32': 'UINT32',
    'uint64': 'UINT64'}


def numpy2blob(tensor: np.ndarray) -> tuple:
    """ Convert the numpy input from user to `Tensor` """
    try:
        dtype = dtype_dict[str(tensor.dtype)]
    except KeyError:
        raise TypeError(f"RedisAI doesn't support tensors of type {tensor.dtype}")
    shape = tensor.shape
    blob = bytes(tensor.data)
    return dtype, shape, blob


def blob2numpy(value: ByteString, shape: Union[list, tuple], dtype: str) -> np.ndarray:
    """ Convert `BLOB` result from RedisAI to `np.ndarray` """
    mm = {
        'FLOAT': 'float32',
        'DOUBLE': 'float64'
    }
    dtype = mm.get(dtype, dtype.lower())
    a = np.frombuffer(value, dtype=dtype)
    return a.reshape(shape)


def list2dict(lst):
    if len(lst) % 2 != 0:
        raise RuntimeError("Can't unpack the list: {}".format(lst))
    out = {}
    for i in range(0, len(lst), 2):
        key = lst[i].decode().lower()
        val = lst[i + 1]
        if key != 'blob' and isinstance(val, bytes):
            val = val.decode()
        out[key] = val
    return out


def recursive_bytetransform(arr: List[AnyStr], target: Callable) -> list:
    """
    Recurse value, replacing each element of b'' with the appropriate element.
    Function returns the same array after inplace operation which updates `arr`

    :param target: Type of tensor | array
    :param arr: The array with b'' numbers or recursive array of b''
    """
    for ix in range(len(arr)):
        obj = arr[ix]
        if isinstance(obj, list):
            recursive_bytetransform(obj, target)
        else:
            arr[ix] = target(obj)
    return arr


def listify(inp: Union[str, Sequence[str]]) -> Sequence[str]:
    return (inp,) if not isinstance(inp, (list, tuple)) else inp


def tensorget_postprocessor(as_numpy, meta_only, rai_result):
    rai_result = list2dict(rai_result)
    if meta_only:
        return rai_result
    elif as_numpy is True:
        return blob2numpy(rai_result['blob'], rai_result['shape'], rai_result['dtype'])
    else:
        target = float if rai_result['dtype'] in ('FLOAT', 'DOUBLE') else int
        recursive_bytetransform(rai_result['values'], target)
        return rai_result
